from collections import Counter
from cqi import CQiClient
from cqi.models.corpora import Corpus as CQiCorpus
from cqi.models.subcorpora import Subcorpus as CQiSubcorpus
from cqi.models.attributes import (
    PositionalAttribute as CQiPositionalAttribute,
    StructuralAttribute as CQiStructuralAttribute
)
from cqi.status import StatusOk as CQiStatusOk
from flask import session
from typing import Dict, List
import gzip
import json
import math
import os
from app import db
from app.models import Corpus
from .utils import lookups_by_cpos, partial_export_subcorpus, export_subcorpus


CQI_EXTENSION_FUNCTION_NAMES: List[str] = [
    'ext_corpus_update_db',
    'ext_corpus_static_data',
    'ext_corpus_paginate_corpus',
    'ext_cqp_paginate_subcorpus',
    'ext_cqp_partial_export_subcorpus',
    'ext_cqp_export_subcorpus',
]


def ext_corpus_update_db(corpus: str) -> CQiStatusOk:
    cqi_client: CQiClient = session['cqi_over_sio']['cqi_client']
    db_corpus_id: int = session['cqi_over_sio']['db_corpus_id']
    db_corpus: Corpus = Corpus.query.get(db_corpus_id)
    cqi_corpus: CQiCorpus = cqi_client.corpora.get(corpus)
    db_corpus.num_tokens = cqi_corpus.size
    db.session.commit()
    return CQiStatusOk()


def ext_corpus_static_data(corpus: str) -> Dict:
    db_corpus_id: int = session['cqi_over_sio']['db_corpus_id']
    db_corpus: Corpus = Corpus.query.get(db_corpus_id)

    static_data_file_path: str = os.path.join(db_corpus.path, 'cwb', 'static.json.gz')
    if os.path.exists(static_data_file_path):
        with open(static_data_file_path, 'rb') as f:
            return f.read()

    cqi_client: CQiClient = session['cqi_over_sio']['cqi_client']
    cqi_corpus: CQiCorpus = cqi_client.corpora.get(corpus)
    cqi_p_attrs: List[CQiPositionalAttribute] = cqi_corpus.positional_attributes.list()
    cqi_s_attrs: List[CQiStructuralAttribute] = cqi_corpus.structural_attributes.list()

    static_data = {
        'corpus': {
            'bounds': [0, cqi_corpus.size - 1],
            'freqs': {}
        },
        'p_attrs': {},
        's_attrs': {},
        'values': {'p_attrs': {}, 's_attrs': {}}
    }

    for p_attr in cqi_p_attrs:
        print(f'corpus.freqs.{p_attr.name}')
        static_data['corpus']['freqs'][p_attr.name] = []
        p_attr_id_list: List[int] = list(range(p_attr.lexicon_size))
        static_data['corpus']['freqs'][p_attr.name].extend(p_attr.freqs_by_ids(p_attr_id_list))
        del p_attr_id_list

        print(f'p_attrs.{p_attr.name}')
        static_data['p_attrs'][p_attr.name] = []
        cpos_list: List[int] = list(range(cqi_corpus.size))
        static_data['p_attrs'][p_attr.name].extend(p_attr.ids_by_cpos(cpos_list))
        del cpos_list

        print(f'values.p_attrs.{p_attr.name}')
        static_data['values']['p_attrs'][p_attr.name] = []
        p_attr_id_list: List[int] = list(range(p_attr.lexicon_size))
        static_data['values']['p_attrs'][p_attr.name].extend(p_attr.values_by_ids(p_attr_id_list))
        del p_attr_id_list

    for s_attr in cqi_s_attrs:
        if s_attr.has_values:
            continue

        static_data['s_attrs'][s_attr.name] = {'lexicon': [], 'values': None}

        if s_attr.name in ['s', 'ent']:
            ##############################################################
            # A faster way to get cpos boundaries for smaller s_attrs    #
            # Note: Needs more testing, don't use it in production       #
            ##############################################################
            cqi_corpus.query('Last', f'<{s_attr.name}> []* </{s_attr.name}>;')
            cqi_subcorpus: CQiSubcorpus = cqi_corpus.subcorpora.get('Last')
            first_match: int = 0
            last_match: int = cqi_subcorpus.size - 1
            match_boundaries = zip(
                range(first_match, last_match + 1),
                cqi_subcorpus.dump(
                    cqi_subcorpus.fields['match'],
                    first_match,
                    last_match
                ),
                cqi_subcorpus.dump(
                    cqi_subcorpus.fields['matchend'],
                    first_match,
                    last_match
                )
            )
            cqi_subcorpus.drop()
            del cqi_subcorpus, first_match, last_match
            for id, lbound, rbound in match_boundaries:
                static_data['s_attrs'][s_attr.name]['lexicon'].append({})
                print(f's_attrs.{s_attr.name}.lexicon.{id}.bounds')
                static_data['s_attrs'][s_attr.name]['lexicon'][id]['bounds'] = [lbound, rbound]
            del match_boundaries

        if s_attr.name != 'text':
            continue

        for id in range(0, s_attr.size):
            static_data['s_attrs'][s_attr.name]['lexicon'].append({})
            # This is a very slow operation, thats why we only use it for
            # the text attribute
            lbound, rbound = s_attr.cpos_by_id(id)
            print(f's_attrs.{s_attr.name}.lexicon.{id}.bounds')
            static_data['s_attrs'][s_attr.name]['lexicon'][id]['bounds'] = [lbound, rbound]
            static_data['s_attrs'][s_attr.name]['lexicon'][id]['freqs'] = {}
            cpos_list: List[int] = list(range(lbound, rbound + 1))
            for p_attr in cqi_p_attrs:
                p_attr_ids: List[int] = []
                p_attr_ids.extend(p_attr.ids_by_cpos(cpos_list))
                print(f's_attrs.{s_attr.name}.lexicon.{id}.freqs.{p_attr.name}')
                static_data['s_attrs'][s_attr.name]['lexicon'][id]['freqs'][p_attr.name] = dict(Counter(p_attr_ids))
                del p_attr_ids
            del cpos_list

        sub_s_attrs: List[CQiStructuralAttribute] = cqi_corpus.structural_attributes.list(filters={'part_of': s_attr})
        print(f's_attrs.{s_attr.name}.values')
        static_data['s_attrs'][s_attr.name]['values'] = [
            sub_s_attr.name[(len(s_attr.name) + 1):]
            for sub_s_attr in sub_s_attrs
        ]
        s_attr_id_list: List[int] = list(range(s_attr.size))
        sub_s_attr_values: List[str] = []
        for sub_s_attr in sub_s_attrs:
            tmp = []
            tmp.extend(sub_s_attr.values_by_ids(s_attr_id_list))
            sub_s_attr_values.append(tmp)
            del tmp
        del s_attr_id_list
        print(f'values.s_attrs.{s_attr.name}')
        static_data['values']['s_attrs'][s_attr.name] = [
            {
                s_attr_value_name: sub_s_attr_values[s_attr_value_name_idx][s_attr_id]
                for s_attr_value_name_idx, s_attr_value_name in enumerate(
                    static_data['s_attrs'][s_attr.name]['values']
                )
            } for s_attr_id in range(0, s_attr.size)
        ]
        del sub_s_attr_values
    print('Saving static data to file')
    with gzip.open(static_data_file_path, 'wt') as f:
        json.dump(static_data, f)
    del static_data
    print('Sending static data to client')
    with open(static_data_file_path, 'rb') as f:
        return f.read()


def ext_corpus_paginate_corpus(
    corpus: str,
    page: int = 1,
    per_page: int = 20
) -> Dict:
    cqi_client: CQiClient = session['cqi_over_sio']['cqi_client']
    cqi_corpus = cqi_client.corpora.get(corpus)
    # Sanity checks
    if (
        per_page < 1
        or page < 1
        or (
            cqi_corpus.size > 0
            and page > math.ceil(cqi_corpus.size / per_page)
        )
    ):
        return {'code': 416, 'msg': 'Range Not Satisfiable'}
    first_cpos = (page - 1) * per_page
    last_cpos = min(cqi_corpus.size, first_cpos + per_page)
    cpos_list = [*range(first_cpos, last_cpos)]
    lookups = lookups_by_cpos(cqi_corpus, cpos_list)
    payload = {}
    # the items for the current page
    payload['items'] = [cpos_list]
    # the lookups for the items
    payload['lookups'] = lookups
    # the total number of items matching the query
    payload['total'] = cqi_corpus.size
    # the number of items to be displayed on a page.
    payload['per_page'] = per_page
    # The total number of pages
    payload['pages'] = math.ceil(payload['total'] / payload['per_page'])
    # the current page number (1 indexed)
    payload['page'] = page if payload['pages'] > 0 else None
    # True if a previous page exists
    payload['has_prev'] = payload['page'] > 1 if payload['page'] else False
    # True if a next page exists.
    payload['has_next'] = payload['page'] < payload['pages'] if payload['page'] else False  # noqa
    # Number of the previous page.
    payload['prev_num'] = payload['page'] - 1 if payload['has_prev'] else None
    # Number of the next page
    payload['next_num'] = payload['page'] + 1 if payload['has_next'] else None
    return payload


def ext_cqp_paginate_subcorpus(
    subcorpus: str,
    context: int = 50,
    page: int = 1,
    per_page: int = 20
) -> Dict:
    corpus_name, subcorpus_name = subcorpus.split(':', 1)
    cqi_client: CQiClient = session['cqi_over_sio']['cqi_client']
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    cqi_subcorpus = cqi_corpus.subcorpora.get(subcorpus_name)
    # Sanity checks
    if (
        per_page < 1
        or page < 1
        or (
            cqi_subcorpus.size > 0
            and page > math.ceil(cqi_subcorpus.size / per_page)
        )
    ):
        return {'code': 416, 'msg': 'Range Not Satisfiable'}
    offset = (page - 1) * per_page
    cutoff = per_page
    cqi_results_export = export_subcorpus(
        cqi_subcorpus, context=context, cutoff=cutoff, offset=offset)
    payload = {}
    # the items for the current page
    payload['items'] = cqi_results_export.pop('matches')
    # the lookups for the items
    payload['lookups'] = cqi_results_export
    # the total number of items matching the query
    payload['total'] = cqi_subcorpus.size
    # the number of items to be displayed on a page.
    payload['per_page'] = per_page
    # The total number of pages
    payload['pages'] = math.ceil(payload['total'] / payload['per_page'])
    # the current page number (1 indexed)
    payload['page'] = page if payload['pages'] > 0 else None
    # True if a previous page exists
    payload['has_prev'] = payload['page'] > 1 if payload['page'] else False
    # True if a next page exists.
    payload['has_next'] = payload['page'] < payload['pages'] if payload['page'] else False  # noqa
    # Number of the previous page.
    payload['prev_num'] = payload['page'] - 1 if payload['has_prev'] else None
    # Number of the next page
    payload['next_num'] = payload['page'] + 1 if payload['has_next'] else None
    return payload


def ext_cqp_partial_export_subcorpus(
    subcorpus: str,
    match_id_list: list,
    context: int = 50
) -> Dict:
    corpus_name, subcorpus_name = subcorpus.split(':', 1)
    cqi_client: CQiClient = session['cqi_over_sio']['cqi_client']
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    cqi_subcorpus = cqi_corpus.subcorpora.get(subcorpus_name)
    cqi_subcorpus_partial_export = partial_export_subcorpus(cqi_subcorpus, match_id_list, context=context)
    return cqi_subcorpus_partial_export


def ext_cqp_export_subcorpus(
    subcorpus: str,
    context: int = 50
) -> Dict:
    corpus_name, subcorpus_name = subcorpus.split(':', 1)
    cqi_client: CQiClient = session['cqi_over_sio']['cqi_client']
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    cqi_subcorpus = cqi_corpus.subcorpora.get(subcorpus_name)
    cqi_subcorpus_export = export_subcorpus(cqi_subcorpus, context=context)
    return cqi_subcorpus_export
